from __future__ import annotations

import copy
import importlib
import importlib.util
import os
import sys
import warnings
from pathlib import Path
from typing import Any

from graph_pes.utils.logger import logger

ALLOWED_PACKAGES = {"torch", "graph_pes"}


def is_allowed_for_import(nested_module_name: str) -> bool:
    """
    Check if a nested module is allowed for import.

    Parameters
    ----------
    nested_module_name
        The name of the nested module.

    Returns
    -------
    bool
        Whether the nested module is allowed for import.
    """

    if any(
        nested_module_name.startswith(package) for package in ALLOWED_PACKAGES
    ):
        return True

    user_string = os.environ.get("GRAPH_PES_ALLOW_IMPORT", "")
    if not user_string:
        return False
    user_specified = set(user_string.split(","))
    logger.debug(f"User specified allowed packages: {user_specified}")
    return any(
        nested_module_name.startswith(package) for package in user_specified
    )


def create_from_data(thing: Any, type_msg: str | None = None) -> Any:
    """
    Create python objects from a data specification.

    Parameters
    ----------
    thing
        A string or a dictionary.
    type_msg
        A message to display if the type of ``thing`` is not recognized.

    Returns
    -------
    Any
        The created object.

    Example
    -------
    >>> create_from_data("torch.nn.ReLU")
    ReLU()
    """

    if isinstance(thing, str):
        return create_from_string(thing)
    elif isinstance(thing, dict):
        return create_from_dict(thing)
    else:
        if type_msg is None:
            type_msg = f"Expected {thing} to be a string or a dictionary."
        raise TypeError(type_msg)


def _import(thing: str) -> Any:
    """
    Import a module or object from a fully qualified name.

    Example
    -------
    >>> _import("torch.nn.Tanh")
    <class 'torch.nn.modules.activation.Tanh'>
    """

    module_name, obj_name = thing.rsplit(".", 1)
    if not is_allowed_for_import(module_name):
        raise ImportError(
            f"Attempted to import {obj_name} from {module_name}, "
            "which is not allowed for safety reasons. Do you want to "
            "allow this import? If so, set the GRAPH_PES_ALLOW_IMPORT "
            "environment variable to include the package you want to "
            "import from: e.g. `export GRAPH_PES_ALLOW_IMPORT=my_package_1,"
            "my_package_2`."
        )

    dir_command_was_run_from = str(Path(os.getcwd()).resolve())
    if dir_command_was_run_from not in sys.path:
        sys.path.append(dir_command_was_run_from)

    try:
        module = importlib.import_module(module_name)
    except ImportError:
        assumed_file = Path(module_name.replace(".", "/")).with_suffix(".py")
        if not assumed_file.exists():
            raise ImportError(
                f"While attempting to import {obj_name} from {module_name}, "
                "we could not find the module or the file "
                f"belonging to {module_name}."
            ) from None

        logger.debug(
            f"Could not directly import '{module_name}' - "
            f"trying to load it from {assumed_file}"
        )
        spec = importlib.util.spec_from_file_location(
            name=module_name, location=assumed_file
        )
        assert spec is not None
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        assert spec.loader is not None
        spec.loader.exec_module(module)

    return getattr(module, obj_name)


def warn_about_import_error(s: str):
    if not Path(s).exists():
        warnings.warn(
            f"Encountered a string ({s}) that looks like it "
            "could be meant to be imported - we couldn't do "
            "this. This may cause issues later.",
            stacklevel=2,
        )


def create_from_dict(d: dict[str, Any]) -> dict[str, Any] | Any:
    """
    Create objects from a dictionary.

    Two cases:
    1. ``d`` has a single key, and the value is a dictionary. In this case, we
       try to import the key (assuming it points to some callable), parse
       the value dictionary as keyword arguments and call the imported object
       with these arguments.
    2. leave the keys of ``d`` as they are, recursively create objects from
       the values and return the resulting dictionary.

    Parameters
    ----------
    d
        A dictionary.

    Returns
    -------
    dict[str, Any] | Any
        The resulting dictionary or object.

    Example
    -------
    A single-key dictionary with a dictionary value gets mapped to an object:

    >>> d = {"torch.nn.Linear": {"in_features": 10, "out_features": 20}}
    >>> create_from_dict(d)
    Linear(in_features=10, out_features=20, bias=True)

    A dictionary with multiple keys gets mapped to a dictionary:
    >>> d = {
    ...     "a": {"b": 1},
    ...     "c": "torch.nn.ReLU()",
    ... }
    >>> create_from_dict(d)
    {'a': {'b': 1}, 'c': ReLU()}

    Note that the string "torch.nn.ReLU()" is imported and called using this
    syntax.
    """
    return _create_from_dict(d, depth=0)


def _create_from_dict(d: dict[str, Any], depth: int) -> dict[str, Any] | Any:
    new_d = copy.deepcopy(d)

    def log(msg):
        logger.debug(f"create_from_dict: {'    ' * depth}{msg}")

    log(f"creating from {d}")

    # 1. recursively create objects from values
    for k, v in new_d.items():
        log(f"{k=}, {v=}")
        if isinstance(v, dict):
            log(f"recursing into {v=}")
            new_d[k] = _create_from_dict(v, depth + 1)
            log(f"received {new_d[k]}")

        elif isinstance(v, str) and looks_importable(v):
            try:
                log(f"trying to import '{v}'")
                obj = create_from_string(v)
                log(f"successfully imported {obj}")
                new_d[k] = obj

            except ImportError:
                log(f"failed to import '{v}'")
                warn_about_import_error(v)

    # 2. if dict has only one key, and that key maps to a dictionary of values
    #    try to import the key, call it with the values as kwargs and
    #    return the result. If any of this fails, return the dict as is.
    if len(new_d) == 1:
        k, kwargs = next(iter(new_d.items()))
        if isinstance(kwargs, dict) and looks_importable(k):
            log(
                "found a single-item dict with importable "
                f"key - trying to import '{k}' and call it with {kwargs}"
            )
            try:
                callable = _import(k)
                result = callable(**kwargs)
                log(f"successfully created {result}")
                log(f"final result: {result}")
                return result
            except ImportError:
                warn_about_import_error(k)

    # 3. if dict has more than one key, return it as is
    log(f"final result: {new_d}")
    return new_d


def looks_importable(s: str):
    return "." in s


def create_from_string(s: str):
    if s.endswith("()"):
        return _import(s[:-2])()
    else:
        return _import(s)



# ruff: noqa: UP006, UP007
# ^^ NB: dacite parsing requires the old type hint syntax in
#        order to be compatible with all versions of Python that
#         we are targeting (3.8+)
from __future__ import annotations

from dataclasses import dataclass
from hashlib import sha256
from typing import Any, Dict, List, Literal, Union

import dacite
import yaml

from graph_pes.data.datasets import FittingData
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.addition import AdditionModel
from graph_pes.training.loss import Loss, TotalLoss
from graph_pes.training.opt import LRScheduler, Optimizer
from graph_pes.training.ptl_utils import VerboseSWACallback

from .utils import create_from_data, create_from_string


@dataclass
class LossSpec:
    """
    Specification for a component of a
    :class:`~graph_pes.training.loss.TotalLoss`.
    """

    component: Union[str, Dict[str, Any]]
    """Point to a :class:`~graph_pes.training.loss.Loss` instance."""

    weight: Union[int, float] = 1.0
    """The weight of this loss component."""


@dataclass
class FittingOptions:
    """Options for the fitting process."""

    pre_fit_model: bool
    """
    Whether to pre-fit the model before training. See
    :meth:`graph_pes.GraphPESModel.pre_fit_all_components` for details.
    """

    max_n_pre_fit: Union[int, None]
    """
    The maximum number of graphs to use for pre-fitting.
    Set to ``None`` to use all the available training data.
    """

    early_stopping_patience: Union[int, None]
    """
    The number of epochs to wait for improvement in the total validation loss
    before stopping training. Set to ``None`` to disable early stopping.
    """

    trainer_kwargs: Dict[str, Any]
    """
    Key-word arguments to pass to the `PyTorch Lightning Trainer 
    <https://lightning.ai/docs/pytorch/stable/common/trainer.html>`__ .
    
    Example
    -------
    .. code-block:: yaml
        
        fitting:
            trainer:
                max_epochs: 10000
                accelerator: gpu
                accumulate_grad_batches: 4
    """

    loader_kwargs: Dict[str, Any]
    """
    Key-word arguments to pass to the underlying 
    :class:`torch.utils.data.DataLoader`.

    See the `PyTorch documentation <https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader>`__ for details.

    Example
    -------
    .. code-block:: yaml

        fitting:    
            loader_kwargs:
                shuffle: true
                seed: 42
                batch_size: 32
                persistent_workers: true
                num_workers: 4
    """  # noqa: E501


@dataclass
class SWAConfig:
    """
    Configuration for Stochastic Weight Averaging.

    Internally, this is handled by `this PyTorch Lightning callback
    <https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.StochasticWeightAveraging.html>`__.
    """

    lr: float
    """
    The learning rate to use during the SWA phase. If not specified,
    the learning rate from the end of the training phase will be used.
    """

    start: Union[int, float] = 0.8
    """
    The epoch at which to start SWA. If a float, it will be interpreted
    as a fraction of the total number of epochs.
    """

    anneal_epochs: int = 10
    """
    The number of epochs over which to linearly anneal the learning rate
    to zero.
    """

    strategy: Literal["linear", "cosine"] = "linear"
    """The strategy to use for annealing the learning rate."""

    def instantiate_lightning_callback(self):
        return VerboseSWACallback(
            swa_lrs=self.lr,
            swa_epoch_start=self.start,
            annealing_epochs=self.anneal_epochs,
            annealing_strategy=self.strategy,
        )


@dataclass
class FittingConfig(FittingOptions):
    """Configuration for the fitting process."""

    optimizer: Union[str, Dict[str, Any]]
    """
    Specification for the optimizer. Point to something that instantiates a
    :class:`~graph_pes.training.opt.Optimizer`.

    Examples
    --------
    The default (see :func:`~graph_pes.training.opt.Optimizer` for details):

    .. code-block:: yaml

        fitting:        
            optimizer:
                graph_pes.training.opt.Optimizer:
                    name: Adam
                    lr: 3e-3
                    weight_decay: 0.0
                    amsgrad: false

    Or a custom one:
    
    .. code-block:: yaml

        fitting:
            optimizer: my.module.MagicOptimizer()
    """

    scheduler: Union[str, Dict[str, Any], None]
    """
    .. _learning rate scheduler:
    
    Specification for the learning rate scheduler. Optional.
    Default is to have no learning rate schedule (``None``).

    Examples
    --------
    .. code-block:: yaml
    
        fitting:
            scheduler:
                graph_pes.training.opt.LRScheduler:
                    name: ReduceLROnPlateau
                factor: 0.5
                patience: 10
    """

    swa: Union[SWAConfig, None]
    """
    Optional, defaults to ``None``.

    .. dropdown:: ``swa`` options

        .. _stochastic weight averaging:

        .. autoclass:: graph_pes.config.config.SWAConfig()
            :members:
    """

    ### Methods ###

    def instantiate_optimizer(self) -> Optimizer:
        return create_from_data(self.optimizer)

    def instantiate_scheduler(self) -> LRScheduler | None:
        if self.scheduler is None:
            return None
        return create_from_data(self.scheduler)


@dataclass
class GeneralConfig:
    """General configuration for a training run."""

    seed: int
    """The global random seed for reproducibility."""

    root_dir: str
    """
    The root directory for this run. 
    
    Results will be stored in ``<root_dir>/<run_id>``, where ``run_id``
    is one of:
    * the user-specified ``run_id`` string
    * a random string generated by ``graph-pes``
    """

    run_id: Union[str, None]
    """A unique identifier for this run."""

    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    """The logging level for the logger."""

    progress: Literal["rich", "logged"]
    """The progress bar style to use."""


@dataclass
class Config:
    """
    A schema for a configuration file to train a
    :class:`~graph_pes.GraphPESModel`.
    """

    model: Union[str, Dict[str, Any]]
    """        
    The model to train.

    .. dropdown:: ``model`` options

        Either point to something that instantiates a
        :class:`~graph_pes.GraphPESModel`...:

        .. code-block:: yaml

            # basic Lennard-Jones model
            model:
                graph_pes.models.LennardJones:
                    sigma: 0.1
                    epsilon: 1.0
            
            # or some custom model
            model: my_model.SpecialModel()
        
        ...or pass a dictionary mapping custom names to
        :class:`~graph_pes.GraphPESModel` instances:

        .. code-block:: yaml
        
            model:
                offset:
                    graph_pes.models.FixedOffset: {H: -123.4, C: -456.7}
                many-body: graph_pes.models.SchNet()

        The latter will instantiate an :class:`~graph_pes.models.AdditionModel`
        with :class:`~graph_pes.models.FixedOffset` and
        :class:`~graph_pes.models.SchNet` components. This is a useful approach
        for dealing with arbitrary offset energies.
    """

    data: Union[str, Dict[str, Any]]
    """
    The data to train on.

    .. dropdown:: ``data`` options
    
        Point to one of the following:

        - a callable that returns a :class:`~graph_pes.data.FittingData` 
          instance
        - a dictionary mapping ``"train"`` and ``"valid"`` keys to callables 
          that return :class:`~graph_pes.data.GraphDataset` instances

        Load custom data from a function with no arguments:

        .. code-block:: yaml
            
            data: my_module.my_fitting_data()

        Point to :func:`graph_pes.data.load_atoms_dataset` with arguments:

        .. code-block:: yaml

            data:
                graph_pes.data.load_atoms_dataset:
                    id: QM9
                    cutoff: 5.0
                    n_train: 10000
                    n_val: 1000
                    property_map:
                        energy: U0

        Point to separate train and validation datasets, taking a random
        1,000 structures from the training file to train from, and all
        structures from the validation file:

        .. code-block:: yaml

            data:
                train:
                    graph_pes.data.file_dataset:
                        path: training_data.xyz
                        cutoff: 5.0
                        n: 1000
                        shuffle: true
                        seed: 42
                valid:
                    graph_pes.data.file_dataset:
                        path: validation_data.xyz
                        cutoff: 5.0
    """

    loss: Union[str, Dict[str, Any], List[LossSpec]]
    """
    The loss function to use.

    .. dropdown:: ``loss`` options

        This config should either point to something that instantiates a
        :class:`~graph_pes.training.loss.Loss`...

        .. code-block:: yaml
            
            # basic per-atom energy loss
            loss: graph_pes.training.loss.PerAtomEnergyLoss()

            # or more fine-grained control
            loss:
                graph_pes.training.loss.Loss:
                    property: energy
                    metric: graph_pes.training.loss.RMSE()

        ...or specify a list of :class:`~graph_pes.config.config.LossSpec`
        instances, each of which points to a 
        :class:`~graph_pes.training.loss.Loss` instance and specifies a weight:

        .. code-block:: yaml
        
            loss:
                - component: graph_pes.training.loss.Loss:
                      property: energy
                      metric: graph_pes.training.loss.RMSE()
                  weight: 1.0
                - component: graph_pes.training.loss.Loss:
                      property: forces
                      metric: graph_pes.training.loss.MAE()
                weight: 10.0
        
        .. autoclass:: graph_pes.config.config.LossSpec()
            :members:
    """

    fitting: FittingConfig
    """
    Extended configuration for the fitting process.

    .. dropdown:: ``fitting`` options

        .. autoclass:: graph_pes.config.config.FittingConfig()
            :members:
            :inherited-members:
    """

    general: GeneralConfig
    """
    General configuration for a training run.

    .. dropdown:: ``general`` options

        .. autoclass:: graph_pes.config.config.GeneralConfig()
            :members:
    """

    wandb: Union[Dict[str, Any], None]
    """
    Configure Weights and Biases logging.

    .. dropdown:: ``wandb`` options

        Disable weights & biases logging:

        .. code-block:: yaml
            
                wandb: null

        Otherwise, provide a dictionary of
        overrides to pass to lightning's `WandbLogger <https://lightning.ai/docs/pytorch/stable/extensions/generated/lightning.pytorch.loggers.WandbLogger.html>`__

        .. code-block:: yaml
        
            wandb:
                project: my_project
                entity: my_entity
                tags: [my_tag]
                
    """  # noqa: E501

    ### Methods ###

    def to_nested_dict(self) -> Dict[str, Any]:
        def _maybe_as_dict(obj):
            if isinstance(obj, list):
                return [_maybe_as_dict(v) for v in obj]
            elif not hasattr(obj, "__dict__"):
                return obj
            return {k: _maybe_as_dict(v) for k, v in obj.__dict__.items()}

        return _maybe_as_dict(self)  # type: ignore

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Config:
        try:
            return dacite.from_dict(
                data_class=cls,
                data=data,
                config=dacite.Config(strict=True),
            )
        except Exception as e:
            raise ValueError(
                "Your configuration file could not be successfully parsed. "
                "Please check that it is formatted correctly. For examples, "
                "please see https://jla-gardner.github.io/graph-pes/cli/graph-pes-train.html"
            ) from e

    def hash(self) -> str:
        """
        Get a unique identifier for this configuration.

        Returns
        -------
        str
            The SHA-256 hash of the configuration.
        """
        return sha256(str(self.to_nested_dict()).encode()).hexdigest()

    def __repr__(self):
        # yaml dump the nested config, complete with defaults,
        # as it would appear in a config.yaml file
        return yaml.dump(self.to_nested_dict(), indent=3, sort_keys=False)

    def instantiate_model(self) -> GraphPESModel:
        obj = create_from_data(self.model)
        if isinstance(obj, GraphPESModel):
            return obj

        elif isinstance(obj, dict):
            all_string_keys = all(isinstance(k, str) for k in obj)
            all_model_values = all(
                isinstance(v, GraphPESModel) for v in obj.values()
            )

            if not all_string_keys or not all_model_values:
                raise ValueError(
                    "Expected a dictionary of named GraphPESModels, "
                    f"but got {obj}."
                )

            try:
                return AdditionModel(**obj)
            except Exception as e:
                raise ValueError(
                    f"Parsed a dictionary, {obj}, from the model config, "
                    "but could not instantiate an AdditionModel from it."
                ) from e

        raise ValueError(
            "Expected to be able to parse a GraphPESModel or a "
            "dictionary of named GraphPESModels from the model config, "
            f"but got something else: {obj}"
        )

    def instantiate_data(self) -> FittingData:
        result: Any = None

        if isinstance(self.data, str):
            result = create_from_string(self.data)

        if isinstance(self.data, dict):
            if len(self.data) == 1:
                result = create_from_data(self.data)
            elif len(self.data) == 2:
                assert self.data.keys() == {"train", "valid"}
                result = FittingData(
                    train=create_from_data(self.data["train"]),
                    valid=create_from_data(self.data["valid"]),
                )

        if result is None:
            raise ValueError(
                "Unexpected data specification. "
                "Please provide a callable or a dictionary containing "
                "a single key (the fully qualified name of some callable) "
                "or two keys ('train' and 'valid') mapping to callables."
            )

        if not isinstance(result, FittingData):
            raise ValueError(
                "Expected to parse a FittingData instance from the data "
                f"config, but got {result}."
            )

        return result

    def instantiate_loss(self) -> TotalLoss:
        if isinstance(self.loss, (str, dict)):
            loss = create_from_data(self.loss)
            if isinstance(loss, Loss):
                return TotalLoss([loss])
            elif isinstance(loss, TotalLoss):
                return loss
            else:
                raise ValueError(
                    "Expected to parse a Loss or TotalLoss instance from the "
                    "loss config, but got something else: {loss}"
                )

        if not all(isinstance(l, LossSpec) for l in self.loss):
            raise ValueError(
                "Expected a list of LossSpec instances from the loss config, "
                f"but got {self.loss}."
            )

        weights = [l.weight for l in self.loss]
        losses = [create_from_data(l.component) for l in self.loss]

        if not all(isinstance(w, (int, float)) for w in weights):
            raise ValueError(
                "Expected a list of weights from the loss config, "
                f"but got {weights}."
            )

        if not all(isinstance(l, Loss) for l in losses):
            raise ValueError(
                "Expected a list of Loss instances from the loss config, "
                f"but got {losses}."
            )

        return TotalLoss(losses, weights)



from __future__ import annotations

from typing import Callable, Sequence

import matplotlib.axes
import matplotlib.lines
import matplotlib.pyplot as plt
import numpy as np
import torch
from ase import Atoms
from cycler import cycler
from matplotlib.ticker import MaxNLocator
from torch import Tensor

from ..atomic_graph import AtomicGraph, PropertyKey, to_batch
from ..graph_pes_model import GraphPESModel

Transform = Callable[[Tensor, AtomicGraph], Tensor]
r"""
Transforms map a property, :math:`x`, to a target property, :math:`y`,
conditioned on an :class:`~graph_pes.AtomicGraph`, :math:`\mathcal{G}`:

.. math::

    T: (x; \mathcal{G}) \mapsto y
"""


def identity(x: Tensor, graph: AtomicGraph) -> Tensor:
    return x


_my_style = {
    "figure.figsize": (3.5, 3),
    "axes.spines.right": False,
    "axes.spines.top": False,
    "legend.frameon": False,
    "legend.fancybox": False,
    "savefig.transparent": False,
    "axes.prop_cycle": cycler(
        "color",
        [
            "#4385be",
            "#ce5d97",
            "#da702c",
            "#8b7ec8",
            "#879a39",
            "#d14d41",
            "#d14d41",
            "#d0a215",
            "#3aa99f",
        ],
    ),
}

plt.rcParams.update(_my_style)


def move_axes(ax: matplotlib.axes.Axes | None = None):  # type: ignore
    """
    Move the axes to the center of the figure
    """
    ax: plt.Axes = ax or plt.gca()
    ax.spines["left"].set_position(("outward", 10))
    ax.spines["bottom"].set_position(("outward", 10))


def parity_plot(
    model: GraphPESModel,
    graphs: AtomicGraph | Sequence[AtomicGraph],
    property: PropertyKey = "energy",
    transform: Transform | None = None,
    units: str | None = None,
    ax: matplotlib.axes.Axes | None = None,  # type: ignore
    **scatter_kwargs,
):
    r"""
    A nicely formatted parity plot of model predictions vs ground truth
    for the given :code:`property`.

    Parameters
    ----------
    model
        The model to for generating predictions.
    graphs
        The graphs to make predictions on.
    property
        The property to plot, e.g. :code:`"energy"`.
    transform
        The transform to apply to the predictions and labels before plotting.
        If not provided, no transform is applied.
    units
        The units of the property, for labelling the axes. If not provided, no
        units are used.
    ax
        The axes to plot on. If not provided, the current axes are used.
    scatter_kwargs
        Keyword arguments to pass to :code:`plt.scatter`.

    Examples
    --------
    Default settings (no units, transforms or custom scatter keywords):

    .. code-block:: python

        parity_plot(model, train, "energy")

    .. image:: Cu-LJ-default-parity.svg
        :align: center

    Custom settings, as seen in
    :doc:`this example notebook <../quickstart/custom-training-loop>`:

    .. code-block:: python

        from graph_pes.atomic_graph import DividePerAtom
        from graph_pes.util import Keys

        for name, data, colour in zip(
            ["Train", "Test"],
            [train, test],
            ["royalblue", "crimson"],
        ):
            parity_plot(
                model,
                data,
                "energy",
                transform=DividePerAtom(),
                units="eV / atom",
                label=name,
                c=colour,
            )

        plt.legend(loc="upper left", fancybox=False);

    .. image:: ../quickstart/parity-plot.svg
        :align: center
    """
    # deal with defaults
    transform = transform or identity

    # get the predictions and labels
    if not isinstance(graphs, AtomicGraph):
        graphs = to_batch(graphs)

    raw_ground_truth = graphs.properties.get(property, None)
    if raw_ground_truth is None:
        raise ValueError(f"Property {property} not in graphs")
    ground_truth = transform(raw_ground_truth, graphs).detach()

    raw_pred = model.predict(graphs, properties=[property]).get(property, None)
    assert raw_pred is not None
    predictions = transform(raw_pred, graphs).detach()

    # plot
    ax: plt.Axes = ax or plt.gca()

    default_kwargs = dict(lw=0, clip_on=False)
    scatter_kwargs = {**default_kwargs, **scatter_kwargs}
    ax.scatter(ground_truth, predictions, **scatter_kwargs)

    # get a point guaranteed to be on the plot
    z = ground_truth.view(-1)[0].item()
    ax.axline((z, z), slope=1, c="k", ls="--", lw=1)

    # aesthetics
    axis_label = (
        f"{property.capitalize()} ({units})" if units else property.capitalize()
    )
    ax.set_xlabel(f"True {axis_label}")
    ax.set_ylabel(f"Predicted {axis_label}")
    ax.set_aspect("equal", "datalim")
    x0, x1 = ax.get_xlim()
    y0, y1 = ax.get_ylim()
    ax.set_xlim(min(x0, y0), max(x1, y1))
    ax.set_ylim(min(x0, y0), max(x1, y1))
    move_axes(ax)

    # 5 ticks each
    ax.xaxis.set_major_locator(MaxNLocator(5))
    ax.yaxis.set_major_locator(MaxNLocator(5))


def dimer_curve(
    model: GraphPESModel,
    system: str,
    units: str | None = None,
    set_to_zero: bool = True,
    rmin: float = 0.9,
    rmax: float = 5.0,
    ax: matplotlib.axes.Axes | None = None,  # type: ignore
    **plot_kwargs,
) -> matplotlib.lines.Line2D:
    r"""
    A nicely formatted dimer curve plot for the given :code:`system`.

    Parameters
    ----------
    model
        The model to for generating predictions.
    system
        The dimer system. Should be one of: a single element, e.g. :code:`"Cu"`,
        or a pair of elements, e.g. :code:`"CuO"`.
    units
        The units of the energy, for labelling the axes. If not provided, no
        units are used.
    set_to_zero
        Whether to set the energy of the dimer at :code:`rmax` to be zero.
    rmin
        The minimum seperation to consider.
    rmax
        The maximum seperation to consider.
    ax
        The axes to plot on. If not provided, the current axes are used.
    plot_kwargs
        Keyword arguments to pass to :code:`plt.plot`.

    Examples
    --------

    .. code-block:: python

        from graph_pes.utils.analysis import dimer_curve
        from graph_pes.models import LennardJones

        dimer_curve(LennardJones(sigma=1.3, epsilon=0.5), system="OH", units="eV")

    .. image:: dimer-curve.svg
        :align: center
    """  # noqa: E501

    trial_atoms = Atoms(system)
    if len(trial_atoms) != 2:
        system = system + "2"

    rs = np.linspace(rmin, rmax, 200)
    dimers = [Atoms(system, positions=[[0, 0, 0], [r, 0, 0]]) for r in rs]
    graphs = [AtomicGraph.from_ase(d, cutoff=rmax + 0.1) for d in dimers]
    batch = to_batch(graphs)

    with torch.no_grad():
        energy = model.predict_energy(batch).numpy()

    if set_to_zero:
        energy -= energy[-1]

    ax: plt.Axes = ax or plt.gca()

    default_kwargs = dict(lw=1, c="k")
    plot_kwargs = {**default_kwargs, **plot_kwargs}
    line = ax.plot(rs, energy, **plot_kwargs)[0]
    assert isinstance(line, matplotlib.lines.Line2D)

    limiting_energy = energy[-1]
    if (energy[:-1] < limiting_energy).any():
        well_depth = limiting_energy - energy[:-1].min()
    else:
        well_depth = 0.1
    bottom = limiting_energy - well_depth * 1.1
    top = limiting_energy + well_depth * 1.1
    ax.set_ylim(bottom, top)
    ax.set_xlabel("r (Å)")
    ax.set_ylabel(f"Dimer Energy ({units})" if units else "Dimer Energy")

    first_in_view = np.where(energy < top)[0][0]
    ax.set_xlim(rs[first_in_view].item() - 0.2, rs[-1] + 0.2)
    move_axes(ax)

    # 5 ticks each
    ax.xaxis.set_major_locator(MaxNLocator(5))
    ax.yaxis.set_major_locator(MaxNLocator(5))

    return line



from __future__ import annotations

import ase
import numpy
from ase.calculators.calculator import Calculator, all_changes
from ase.stress import full_3x3_to_voigt_6_stress
from graph_pes.atomic_graph import (
    AtomicGraph,
    PropertyKey,
    number_of_atoms,
    to_batch,
)
from graph_pes.graph_pes_model import GraphPESModel


class GraphPESCalculator(Calculator):
    """
    ASE calculator wrapping any GraphPESModel.

    Parameters
    ----------
    model
        The model to use for the calculation.
    device
        The device to use for the calculation, e.g. "cpu" or "cuda".
    **kwargs
        Properties passed to the :class:`ase.calculators.calculator.Calculator`
        base class.
    """

    implemented_properties = ["energy", "forces", "stress"]

    def __init__(
        self,
        model: GraphPESModel,
        device: str = "cpu",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = model.to(device)
        self.model.eval()
        self.device = device

    def calculate(
        self,
        atoms: ase.Atoms | None = None,
        properties: list[str] | None = None,
        system_changes: list[str] = all_changes,
    ):
        if properties is None:
            properties = ["energy"]

        # call to base-class to set atoms attribute
        super().calculate(atoms)
        assert self.atoms is not None and isinstance(self.atoms, ase.Atoms)

        # account for numerical inprecision by nudging the cutoff up slightly
        graph = AtomicGraph.from_ase(
            self.atoms, self.model.cutoff.item() + 0.001
        )
        graph = graph.to(self.device)

        results = {
            k: v.detach().cpu().numpy()
            for k, v in self.model.predict(
                graph,
                properties=properties,  # type: ignore
            ).items()
            if k in properties
        }
        if "energy" in properties:
            results["energy"] = results["energy"].item()
        if "stress" in properties:
            results["stress"] = full_3x3_to_voigt_6_stress(results["stress"])

        self.results = results

    def batched_prediction(
        self,
        structures: list[AtomicGraph | ase.Atoms],
        properties: list[PropertyKey] | None = None,
        batch_size: int = 5,
    ) -> list[dict[str, numpy.ndarray]]:
        """
        Make a batched prediction on the given list of structures.

        Parameters
        ----------
        structures
            A list of :class:`~graph_pes.AtomicGraph` or
            :class:`ase.Atoms` objects.
        properties
            The properties to predict.
        batch_size
            The number of structures to predict at once.

        Examples
        --------
        >>> calculator = GraphPESCalculator(model, device="cuda")
        >>> structures = [Atoms(...), Atoms(...), Atoms(...)]
        >>> predictions = calculator.batched_prediction(
        ...     structures,
        ...     properties=["energy", "forces"],
        ...     batch_size=2,
        ... )
        >>> print(predictions)
        [{'energy': array(...), 'forces': array(...)},
         {'energy': array(...), 'forces': array(...)},
         {'energy': array(...), 'forces': array(...)}]
        """
        # recursive batching
        if len(structures) > batch_size:
            first_batch, rest = structures[:batch_size], structures[batch_size:]
            return self.batched_prediction(
                first_batch, properties, batch_size
            ) + self.batched_prediction(rest, properties, batch_size)

        if properties is None:
            properties = ["energy", "forces", "stress"]
        graphs = [
            AtomicGraph.from_ase(s, self.model.cutoff.item() + 0.001)
            if isinstance(s, ase.Atoms)
            else s
            for s in structures
        ]
        batch_preds = self.model.predict(
            to_batch(graphs), properties=properties
        )

        # de-convolve the batch predictions into the original structures
        preds_list = []
        atom_index = 0
        for structure_index, graph in enumerate(graphs):
            N = number_of_atoms(graph)
            preds = {}
            if "energy" in properties:
                preds["energy"] = batch_preds["energy"][structure_index]
            if "forces" in properties:
                preds["forces"] = batch_preds["forces"][
                    atom_index : atom_index + N
                ]
            if "stress" in properties:
                preds["stress"] = batch_preds["stress"][structure_index]
            if "local_energies" in properties:
                preds["local_energies"] = batch_preds["local_energies"][
                    atom_index : atom_index + N
                ]
            preds_list.append(preds)
            atom_index += N
        return preds_list



from __future__ import annotations

from pathlib import Path
from typing import Literal

import pytorch_lightning as pl
import torch
from graph_pes.atomic_graph import (
    AtomicGraph,
    PropertyKey,
    number_of_structures,
)
from graph_pes.config import FittingOptions
from graph_pes.data.datasets import FittingData
from graph_pes.data.loader import GraphDataLoader
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.training.loss import RMSE, Loss, PerAtomEnergyLoss, TotalLoss
from graph_pes.training.opt import LRScheduler, Optimizer
from graph_pes.training.ptl_utils import LoggedProgressBar, ModelTimer
from graph_pes.training.util import log_model_info, sanity_check
from graph_pes.utils.logger import logger
from pytorch_lightning.callbacks import (
    EarlyStopping,
    LearningRateMonitor,
    ModelCheckpoint,
    ProgressBar,
    RichProgressBar,
)
from pytorch_lightning.loggers import Logger
from pytorch_lightning.utilities.types import OptimizerLRSchedulerConfig

VALIDATION_LOSS_KEY = "valid/loss/total"


def train_with_lightning(
    trainer: pl.Trainer,
    model: GraphPESModel,
    data: FittingData,
    loss: TotalLoss,
    fit_config: FittingOptions,
    optimizer: Optimizer,
    scheduler: LRScheduler | None = None,
):
    # - prepare the data
    if trainer.global_rank == 0:
        logger.info("Preparing data")
        data.train.prepare_data()
        data.valid.prepare_data()
    trainer.strategy.barrier("data prepare")

    logger.info("Setting up datasets")
    data.train.setup()
    data.valid.setup()

    loader_kwargs = {**fit_config.loader_kwargs}
    loader_kwargs["shuffle"] = True
    train_loader = GraphDataLoader(data.train, **loader_kwargs)
    loader_kwargs["shuffle"] = False
    valid_loader = GraphDataLoader(data.valid, **loader_kwargs)

    # - maybe do some pre-fitting
    if trainer.global_rank == 0 and fit_config.pre_fit_model:
        pre_fit_dataset = data.train
        if fit_config.max_n_pre_fit is not None:
            pre_fit_dataset = pre_fit_dataset.sample(fit_config.max_n_pre_fit)
        logger.info(
            f"Pre-fitting the model on {len(pre_fit_dataset):,} samples"
        )
        model.pre_fit_all_components(pre_fit_dataset)
    trainer.strategy.barrier("pre-fit")

    # - log the model info
    if trainer.global_rank == 0:
        log_model_info(model, trainer.logger)

    # - sanity checks
    sanity_check(model, next(iter(train_loader)))

    # - create the task (a pytorch lightning module)
    task = LearnThePES(model, loss, optimizer, scheduler)

    # - train the model
    error = None
    try:
        trainer.fit(task, train_loader, valid_loader)
    except Exception as e:
        logger.error(f"Training failed: {e}")
        error = e

    # - load the best weights
    try:
        task.load_best_weights(model, trainer)
    except Exception as e:
        logger.error(f"Failed to load best weights: {e}")
        if error is None:
            error = e
        raise error from None


class LearnThePES(pl.LightningModule):
    def __init__(
        self,
        model: GraphPESModel,
        loss: TotalLoss,
        optimizer: Optimizer,
        scheduler: LRScheduler | None,
    ):
        super().__init__()
        self.model = model
        self.optimizer_factory = optimizer
        self.scheduler_factory = scheduler
        self.total_loss = loss
        self.properties: list[PropertyKey] = [
            component.property for component in self.total_loss.losses
        ]

        # TODO: we want to base this on actually available data
        # not just the total loss components
        validation_metrics = []

        existing_loss_names = [l.name for l in self.total_loss.losses]

        if "energy" in self.properties:
            pael = PerAtomEnergyLoss()
            if pael.name not in existing_loss_names:
                validation_metrics.append(PerAtomEnergyLoss())

        if "forces" in self.properties:
            fr = Loss("forces", RMSE())
            if fr.name not in existing_loss_names:
                validation_metrics.append(fr)

        self.validation_metrics = validation_metrics

    def forward(self, graphs: AtomicGraph) -> torch.Tensor:
        """Get the energy"""
        return self.model.predict_energy(graphs)

    def _step(self, graph: AtomicGraph, prefix: Literal["train", "valid"]):
        """
        Get (and log) the losses for a training/validation step.
        """

        def log(name: str, value: torch.Tensor | float):
            if isinstance(value, torch.Tensor):
                value = value.item()

            return self.log(
                f"{prefix}/{name}",
                value,
                prog_bar=prefix == "valid" and "metric" in name,
                on_step=prefix == "train",
                on_epoch=prefix == "valid",
                sync_dist=prefix == "valid",
                batch_size=number_of_structures(graph),
            )

        # generate prediction:
        predictions = self.model.predict(graph, properties=self.properties)

        # compute the loss and its sub-components
        total_loss_result = self.total_loss(predictions, graph)

        # log
        log("loss/total", total_loss_result.loss_value)
        for name, loss_pair in total_loss_result.components.items():
            log(f"metrics/{name}", loss_pair.loss_value)
            log(f"loss/{name}_component", loss_pair.weighted_loss_value)

        # log additional values during validation
        if prefix == "valid":
            with torch.no_grad():
                for val_loss in self.validation_metrics:
                    value = val_loss(predictions, graph)
                    log(f"metrics/{val_loss.name}", value)

        return total_loss_result.loss_value

    def training_step(self, structure: AtomicGraph, _):
        return self._step(structure, "train")

    def validation_step(self, structure: AtomicGraph, _):
        return self._step(structure, "valid")

    def configure_optimizers(
        self,
    ) -> torch.optim.Optimizer | OptimizerLRSchedulerConfig:
        opt = self.optimizer_factory(self.model)
        if self.scheduler_factory is None:
            return opt

        scheduler = self.scheduler_factory(opt)

        if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
            trainer = self.trainer
            assert trainer is not None

            # if the user is validating once every n training epochs,
            # set the scheduler to update every n training epochs
            if trainer.check_val_every_n_epoch:
                config = {
                    "interval": "epoch",  # really this means training epoch
                    "frequency": trainer.check_val_every_n_epoch,
                    "monitor": VALIDATION_LOSS_KEY,
                    "strict": True,
                }

            # otherwise, they are validating every n training steps:
            # set the scheduler to update every n training steps
            else:
                if isinstance(trainer.val_check_interval, float):
                    raise ValueError(
                        "`graph-pes` does not support specifying "
                        "`val_check_interval` as a fraction of the training "
                        "data size. Please specify it as an integer number of "
                        "training steps."
                    )
                config = {
                    "interval": "step",
                    "frequency": trainer.val_check_interval,
                    "monitor": VALIDATION_LOSS_KEY,
                    "strict": True,
                }

        else:
            # vanilla LR scheduler: update every training step
            config = {
                "interval": "step",
                "frequency": 1,
            }

        logger.debug(f"Using LR scheduler config:\n{config}")
        config["scheduler"] = scheduler
        return {"optimizer": opt, "lr_scheduler": config}  # type: ignore

    def on_validation_model_eval(self, *args, **kwargs):
        super().on_validation_model_eval(*args, **kwargs)
        # we override the defaults to turn on gradient tracking for the
        # validation step since we (might) compute the forces using autograd
        torch.set_grad_enabled(True)

    @classmethod
    def load_best_weights(
        cls,
        model: GraphPESModel,
        trainer: pl.Trainer | None = None,
        checkpoint_path: Path | str | None = None,
    ):
        if checkpoint_path is None and trainer is None:
            raise ValueError(
                "Either trainer or checkpoint_path must be provided"
            )
        if checkpoint_path is None:
            path = trainer.checkpoint_callback.best_model_path  # type: ignore
        else:
            path = Path(checkpoint_path)
        logger.info(f'Loading best weights from "{path}"')
        checkpoint = torch.load(path)
        state_dict = {
            k.replace("model.", "", 1): v
            for k, v in checkpoint["state_dict"].items()
            if k.startswith("model.")
        }
        model.load_state_dict(state_dict)


def create_trainer(
    early_stopping_patience: int | None = None,
    kwarg_overloads: dict | None = None,
    valid_available: bool = False,
    logger: Logger | None = None,
    output_dir: Path | None = None,
    progress: Literal["rich", "logged"] = "rich",
) -> pl.Trainer:
    # create the default callbacks
    callbacks: dict[str, pl.Callback] = dict(
        lr=LearningRateMonitor(logging_interval="epoch"),
        checkpoint=ModelCheckpoint(
            dirpath=None if not output_dir else output_dir / "checkpoints",
            monitor=VALIDATION_LOSS_KEY if valid_available else None,
            filename="best",
            mode="min",
            save_top_k=1,
            save_weights_only=True,
        ),
        progress_bar={
            "rich": RichProgressBar(),
            "logged": LoggedProgressBar(),
        }[progress],
        timer=ModelTimer(),
    )
    if early_stopping_patience is not None:
        if not valid_available:
            raise ValueError(
                "Early stopping requires validation data to be available"
            )
        callbacks["early_stopping"] = EarlyStopping(
            monitor=VALIDATION_LOSS_KEY,
            patience=early_stopping_patience,
            mode="min",
            min_delta=1e-6,
        )

    # find any user defined callbacks ...
    overloads = kwarg_overloads or {}
    overloaded_callbacks = overloads.pop("callbacks", [])

    # ... and overwrite the default callbacks where necessary
    for cb in overloaded_callbacks:
        # we don't want two progress bars: use the non-default one
        if isinstance(cb, ProgressBar):
            callbacks["progress_bar"] = cb
        # all other callbacks are just added on as extras
        else:
            callbacks[str(cb)] = cb

    return pl.Trainer(
        **overloads,
        logger=logger,
        callbacks=list(callbacks.values()),
    )



from __future__ import annotations

from typing import Sequence

from graph_pes.atomic_graph import (
    AtomicGraph,
    PropertyKey,
    has_cell,
    number_of_atoms,
    number_of_structures,
    to_batch,
)
from graph_pes.config import FittingOptions
from graph_pes.data.datasets import (
    FittingData,
    GraphDataset,
    SequenceDataset,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.training.loss import RMSE, Loss, TotalLoss
from graph_pes.training.opt import Optimizer
from graph_pes.training.ptl import create_trainer, train_with_lightning
from graph_pes.utils.logger import logger


# TODO: add logger folder
def train_the_model(
    model: GraphPESModel,
    train_data: Sequence[AtomicGraph] | GraphDataset,
    val_data: Sequence[AtomicGraph] | GraphDataset,
    optimizer: Optimizer | None = None,  # TODO change to factory
    loss: TotalLoss | Loss | None = None,
    *,
    config_to_log: dict | None = None,
    batch_size: int = 32,
    pre_fit_model: bool = True,
    max_n_pre_fit: int | None = 5_000,
    # pytorch lightning
    trainer_options: dict | None = None,
):
    """
    Train a model on the given data.
    TODO: fill this out
    Simple.
    """

    # defaults:
    if trainer_options is None:
        trainer_options = {}
    if config_to_log is None:
        config_to_log = {}
    if optimizer is None:
        optimizer = Optimizer("Adam", lr=3e-3)

    # create data module
    if not isinstance(train_data, GraphDataset):
        train_data = SequenceDataset(train_data)
    if not isinstance(val_data, GraphDataset):
        val_data = SequenceDataset(val_data)

    # sanity check on cpu TODO: move to ptl
    dummy_batch = to_batch(list(train_data.up_to_n(10)))
    try:
        model.get_all_PES_predictions(dummy_batch)
    except Exception as e:
        raise ValueError("The model does not appear to work") from e

    # process and validate the loss
    logger.debug(f"Available properties: {train_data.properties}")
    total_loss = process_loss(loss, train_data.properties)

    training_on: list[PropertyKey] = [
        component.property for component in total_loss.losses
    ]
    for prop in training_on:
        if prop not in train_data.properties:
            raise ValueError(
                f"Can't train on {prop} without the corresponding labels. "
                f"Available labels: {train_data.properties}"
            )
    if not training_on:
        raise ValueError("Found no properties to train on")

    logger.info(f"{total_loss}")

    # sanity checking the data
    expected_shapes = {
        "energy": (number_of_structures(dummy_batch),),
        "forces": (number_of_atoms(dummy_batch), 3),
        "stress": (number_of_structures(dummy_batch), 3, 3),
    }
    for prop in training_on:
        if dummy_batch.properties[prop].shape != expected_shapes[prop]:
            raise ValueError(
                f"Expected {prop} to have shape {expected_shapes[prop]}, "
                f"but found {dummy_batch.properties[prop].shape}"
            )
    if "stress" in training_on and not has_cell(dummy_batch):
        raise ValueError("Can't train on stress without cell information.")

    # train the model
    options = FittingOptions(
        pre_fit_model=pre_fit_model,
        max_n_pre_fit=max_n_pre_fit,
        early_stopping_patience=100,
        trainer_kwargs=trainer_options,
        loader_kwargs=dict(batch_size=batch_size),
    )
    trainer = create_trainer(
        early_stopping_patience=100,
        valid_available=True,
        kwarg_overloads=trainer_options,
    )
    train_with_lightning(
        trainer,
        model,
        data=FittingData(train_data, val_data),
        loss=total_loss,
        fit_config=options,
        optimizer=optimizer,
        scheduler=None,
    )


def process_loss(
    loss: TotalLoss | Loss | None,
    available_properties: list[PropertyKey],
) -> TotalLoss:
    if isinstance(loss, TotalLoss):
        return loss

    elif isinstance(loss, Loss):
        return TotalLoss([loss], [1.0])

    if loss is not None:
        raise ValueError(
            "Invalid loss: must be a TotalLoss, a Loss, or None. "
            f"Got {type(loss)}"
        )

    return TotalLoss(
        losses=[Loss(key, metric=RMSE()) for key in available_properties],
        weights=None,
    )



import warnings
from typing import (
    TYPE_CHECKING,
    Dict,
    Final,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Sequence,
    Union,
    cast,
)

import ase
import numpy as np
import torch
from ase.neighborlist import neighbor_list
from typing_extensions import TypeAlias

from graph_pes.utils.misc import (
    all_equal,
    is_being_documented,
    left_aligned_div,
    left_aligned_mul,
    to_significant_figures,
    uniform_repr,
)

DEFAULT_CUTOFF: Final[float] = 5.0


PropertyKey: TypeAlias = Literal["local_energies", "forces", "energy", "stress"]
ALL_PROPERTY_KEYS: Final[List[PropertyKey]] = [
    "local_energies",
    "forces",
    "energy",
    "stress",
]

if not TYPE_CHECKING and not is_being_documented():
    # torchscript doesn't handle TypedDicts or Literal types:
    # at run-time, we just use less specific, but still correct, types
    Properties: TypeAlias = Dict[str, torch.Tensor]  # noqa: UP006 <- torchscript issue
    PropertyKey: TypeAlias = str


class AtomicGraph(NamedTuple):
    # special members of other: cutoff, batch, ptr
    r"""
    An :class:`AtomicGraph` represents an atomic structure.
    Each node corresponds to an atom, and each directed edge links a central
    atom to a "bonded" neighbour.

    We implement such graphs as (immutable) :class:`~typing.NamedTuple`\ s.
    This allows for easy serialisation and compatibility with PyTorch,
    TorchScript and other libraries. These objects, and all functions that
    operate on them, are compatible with both isolated and periodic structures.

    Batches of multiple structures are represented by a single
    :class:`AtomicGraph` object containing multiple disjoint subgraphs. See
    :func:`~graph_pes.atomic_graph.to_batch` for more information.

    Properties
    ++++++++++

    Below we assume the graph contains ``N`` atoms, ``E`` edges (and ``S``
    structures if the graph is batched).

    We use two examples to illustrate the various properties:


    .. dropdown:: Water molecule

        .. code-block:: python

            >>> from ase.build import molecule
            >>> from load_atoms import view
            >>> water = molecule("H2O")
            >>> view(water, show_bonds=True)

        .. raw:: html
            :file: ../../docs/source/_static/water.html

        We can see that there are 3 atoms, with 2 bonds (and therefore 4
        directed edges):

        .. code-block:: python

            >>> from graph_pes import AtomicGraph
            >>> water_graph = AtomicGraph.from_ase(water, cutoff=1.2)
            >>> water_graph
            AtomicGraph(atoms=3, edges=4, has_cell=False, cutoff=1.2)

    .. dropdown:: Sodium crystal

        .. code-block:: python

            >>> from ase.build import bulk
            >>> from load_atoms import view
            >>> sodium = bulk("Na")
            >>> view(sodium.repeat(3), show_bonds=True)

        .. raw:: html
            :file: ../../docs/source/_static/Na.html

        This structure has a single atom within a periodic cell. If you
        look closely, you can see that this atom has 8 nearest neighbours.
        Only "source" atoms within the unit cell are included in the
        neighbour list, and hence there are 8 edges:

        .. code-block:: python

            >>> sodium_graph = AtomicGraph.from_ase(sodium, cutoff=3.7)
            >>> sodium_graph
            AtomicGraph(atoms=1, edges=8, has_cell=True, cutoff=3.7)
    """

    Z: torch.Tensor
    """
    The atomic numbers of the atoms in the graph, of shape ``(N,)``.

    .. code-block:: python

        >>> water_graph.Z
        tensor([8, 1, 1])
    
    .. code-block:: python

        >>> sodium_graph.Z
        tensor([11])
    """

    R: torch.Tensor
    """
    The cartesian positions of the atoms in the graph, of shape ``(N, 3)``.

    .. code-block:: python

        >>> water_graph.R
        tensor([[ 0.0000,  0.0000,  0.1193],
                [ 0.0000,  0.7632, -0.4770],
                [ 0.0000, -0.7632, -0.4770]])
    
    .. code-block:: python
        
        >>> sodium_graph.R
        tensor([[0., 0., 0.]])
    """

    cell: torch.Tensor
    """
    The unit cell vectors, of shape ``(3, 3)`` for a single structure, or
    ``(S, 3, 3)`` for a batched graph. If the structure is non-periodic,
    this will be all zeros.

    .. code-block:: python

        >>> water_graph.cell
        tensor([[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]])
    
    .. code-block:: python
        
        >>> sodium_graph.cell
        tensor([[-2.1150,  2.1150,  2.1150],
                [ 2.1150, -2.1150,  2.1150],
                [ 2.1150,  2.1150, -2.1150]])
    """

    neighbour_list: torch.Tensor
    """
    A neighbour list, of shape ``(2, E)``, where 
    ``i, j = graph.neighbour_list[:, k]`` is the ``k``'th directed edge
    in the graph, linking atom ``i`` to atom ``j``.

    .. code-block:: python

        >>> water_graph.neighbour_list
        tensor([[0, 0, 1, 2],
                [1, 2, 0, 0]])
    
    .. code-block:: python
        
        >>> sodium_graph.neighbour_list
        tensor([[0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0]])
    """

    neighbour_cell_offsets: torch.Tensor
    """
    The offsets of the neighbours of each atom in units of the unit cell
    vectors, of shape ``(E, 3)``, such that:

    .. code-block:: python

        # k-th edge
        i, j = graph.neighbour_list[:, k]
        kth_displacement_vector = (
            graph.R[j] 
            + graph.neighbour_cell_offsets[k] @ graph.cell
            - graph.R[i]
        )

    In the case of an isolated, non-periodic structure, these will be all zeros.

    .. code-block:: python

        >>> water_graph.neighbour_cell_offsets
        tensor([[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]])
    
    .. code-block:: python

        >>> sodium_graph.neighbour_cell_offsets
        tensor([[ 0.,  0.,  1.],
                [ 0.,  1.,  0.],
                [ 1.,  1.,  1.],
                [ 1.,  0.,  0.],
                [-1.,  0.,  0.],
                [-1., -1., -1.],
                [ 0., -1.,  0.],
                [ 0.,  0., -1.]])

    """

    properties: Dict[PropertyKey, torch.Tensor]
    """
    A dictionary containing potential energy surface (PES) related properties
    of the graph.

    .. list-table::
        :header-rows: 1

        * - Key
          - Shape
          - Description
        * - :code:`"local_energies"`
          - :code:`(N,)`
          - contribution to the total energy from each atom
        * - :code:`"energy"`
          - ``()`` 
          
            (``(S,)`` if batched)
          - total energy of the structure
        * - :code:`"forces"`
          - :code:`(N, 3)`
          - force on each atom
        * - :code:`"stress"`
          - ``(3, 3)`` 
          
            (``(S, 3, 3)`` if batched)
          - stress tensor (see :doc:`../theory`)
    """

    other: Dict[str, torch.Tensor]  # noqa: UP006 <- torchscript issue
    """
    A dictionary containing any other additional information about the graph.
    Feel free to populate this as you wish. Internally, we use this to store
    the cutoff distance if the neighbour list was created using a uniform
    cutoff, as well as the ``"batch"`` and ``"ptr"`` arrays used to indicate
    that this graph represents a batch of sub-graphs.
    """

    @classmethod
    def from_ase(
        cls,
        structure: ase.Atoms,
        cutoff: float = DEFAULT_CUTOFF,
        property_mapping: Union[Mapping[str, PropertyKey], None] = None,
        others_to_include: Union[Sequence[str], None] = None,
    ) -> "AtomicGraph":
        r"""
        Convert an :class:`ase.Atoms` object to an :class:`AtomicGraph`.

        Parameters
        ----------
        structure
            The :class:`ase.Atoms` object to convert.
        cutoff
            The cutoff distance for neighbour finding.
        property_mapping
            An optional mapping of the form ``{key_on_structure:
            key_for_graph}`` defining how relevant properties are labelled on
            the :class:`ase.Atoms` object. If not provided, this function will
            extract all of ``"energy"``, ``"forces"``, or ``"stress"`` from the
            ``.info`` and ``.arrays`` dicts if they are present.
        others_to_include
            An optional list of other ``.info``/``.arrays`` keys to include in
            the graph's ``other`` dict. The corresponding values will be
            converted to :class:`torch.Tensor`\ s.

        Example
        -------

        .. code-block:: python

            >>> from ase.build import molecule
            >>> from graph_pes import AtomicGraph

            >>> # create a a structure with some extra info
            >>> atoms = molecule("H2O")
            >>> atoms.info["DFT_energy"] = -10.0
            >>> atoms.info["unique_id"] = 1234

            >>> # default behaviour:
            >>> AtomicGraph.from_ase(atoms)
            AtomicGraph(atoms=3, edges=6, has_cell=False, cutoff=5.0)

            >>> # specify how to map properties, and other things to include
            >>> AtomicGraph.from_ase(
            ...     atoms,
            ...     property_mapping={
            ...         "DFT_energy": "energy",
            ...     },
            ...     others_to_include=["unique_id"],
            ... )
            AtomicGraph(
                atoms=3,
                edges=6,
                has_cell=False,
                cutoff=5.0,
                properties=['energy'],
                other=['unique_id']
            )
        """

        dtype = torch.get_default_dtype()

        # structure
        Z = torch.tensor(structure.numbers, dtype=torch.long)
        R = torch.tensor(structure.positions, dtype=dtype)
        cell = torch.tensor(structure.cell.array, dtype=dtype)

        # neighbour list
        i, j, offsets = neighbor_list("ijS", structure, cutoff)
        neighbour_list = torch.tensor(np.vstack([i, j]), dtype=torch.long)
        neighbour_cell_offsets = torch.tensor(offsets, dtype=dtype)

        # properties
        properties: dict[PropertyKey, torch.Tensor] = {}
        other: dict[str, torch.Tensor] = {"cutoff": torch.tensor(cutoff)}

        if property_mapping is None:
            all_keys = set(structure.info) | set(structure.arrays)
            property_mapping = {
                k: cast(PropertyKey, k)
                for k in ["energy", "forces", "stress"]
                if k in all_keys
            }
        if others_to_include is None:
            others_to_include = []

        for key, value in list(structure.info.items()) + list(
            structure.arrays.items()
        ):
            if key in property_mapping:
                property = property_mapping[key]
                properties[property] = torch.tensor(value, dtype=dtype)
            elif key in others_to_include:
                other[key] = torch.tensor(value, dtype=dtype)

        missing = set(
            structure_key
            for structure_key, graph_key in property_mapping.items()
            if graph_key not in properties
        )
        if missing:
            raise ValueError(f"Unable to find properties: {missing}")

        return cls.create_with_defaults(
            Z=Z,
            R=R,
            cell=cell,
            neighbour_list=neighbour_list,
            neighbour_cell_offsets=neighbour_cell_offsets,
            properties=properties,
            other=other,
        )

    @classmethod
    def create_with_defaults(
        cls,
        Z: torch.Tensor,
        R: torch.Tensor,
        cell: Union[torch.Tensor, None] = None,
        neighbour_list: Union[torch.Tensor, None] = None,
        neighbour_cell_offsets: Union[torch.Tensor, None] = None,
        properties: Union[Dict[PropertyKey, torch.Tensor], None] = None,
        other: Union[Dict[str, torch.Tensor], None] = None,
    ) -> "AtomicGraph":
        """
        Create an :class:`AtomicGraph`, populating missing values with defaults.

        Parameters
        ----------
        Z
            The atomic numbers.
        R
            The cartesian positions.
        cell
            The unit cell. Defaults to ``torch.zeros(3, 3)``.
        neighbour_list
            The neighbour list. Defaults to ``torch.zeros(2, 0)``.
        neighbour_cell_offsets
            The neighbour cell offsets. Defaults to ``torch.zeros(0, 3)``.
        properties
            The properties. Defaults to ``{}``.
        other
            The other information. Defaults to ``{}``.
        """

        if cell is None:
            cell = torch.zeros(3, 3)
        if neighbour_list is None:
            neighbour_list = torch.zeros(2, 0)
        if neighbour_cell_offsets is None:
            neighbour_cell_offsets = torch.zeros(0, 3)
        if properties is None:
            properties = {}
        if other is None:
            other = {}
        return cls(
            Z=Z,
            R=R,
            cell=cell,
            neighbour_list=neighbour_list,
            neighbour_cell_offsets=neighbour_cell_offsets,
            properties=properties,
            other=other,
        )

    def to(self, device: Union[torch.device, str]) -> "AtomicGraph":
        """Move this graph to the specified device."""

        properties: dict[PropertyKey, torch.Tensor] = {
            k: v.to(device) for k, v in self.properties.items()
        }
        return AtomicGraph(
            Z=self.Z.to(device),
            R=self.R.to(device),
            cell=self.cell.to(device),
            neighbour_list=self.neighbour_list.to(device),
            neighbour_cell_offsets=self.neighbour_cell_offsets.to(device),
            properties=properties,
            other={k: v.to(device) for k, v in self.other.items()},
        )

    def __repr__(self):
        info = {}

        if is_batch(self):
            name = "AtomicGraphBatch"
            info["structures"] = self.other["batch"].max().item() + 1
        else:
            name = "AtomicGraph"

        info["atoms"] = number_of_atoms(self)
        info["edges"] = number_of_edges(self)
        info["has_cell"] = has_cell(self)
        if "cutoff" in self.other:
            info["cutoff"] = to_significant_figures(
                self.other["cutoff"].item(), 3
            )
        if self.properties:
            info["properties"] = available_properties(self)
        actual_other = {
            k: v
            for k, v in self.other.items()
            if k not in {"cutoff", "batch", "ptr"}
        }
        if actual_other:
            info["other"] = list(actual_other.keys())

        return uniform_repr(name, **info, indent_width=4)


############################### BATCHING ###############################


def to_batch(
    graphs: Sequence[AtomicGraph],
) -> AtomicGraph:
    """
    Collate a sequence of atomic graphs into a single batch object.

    The ``Z``, ``R``, ``neighbour_list``, and ``neighbour_cell_offsets``
    properties are concatenated along the first axis, while the ``cell``
    property is stacked along a new batch dimension.

    Values in the ``"other"`` dictionary are concatenated along the first axis
    if they appear to be a per-atoms property (i.e. their first dimension
    matches the number of atoms in the structure). Otherwise, they are stacked
    along a new batch dimension.

    Parameters
    ----------
    graphs
        The graphs to collate.

    Examples
    --------

    A basic example:

    >>> from ase.build import molecule
    >>> from graph_pes import AtomicGraph, to_batch
    >>> graphs = [
    ...     AtomicGraph.from_ase(molecule("H2O")),
    ...     AtomicGraph.from_ase(molecule("CH4")),
    ... ]
    >>> batch = to_batch(graphs)
    >>> batch.other["batch"]  # H20 has 3 atoms, CH4 has 5
    tensor([0, 0, 0, 1, 1, 1, 1, 1])
    >>> batch.other["ptr"]  # offset of first atom of each graph
    tensor([0, 3, 8])
    >>> batch.Z.shape
    torch.Size([8])
    >>> batch.R.shape
    torch.Size([8, 3])
    >>> batch.cell.shape
    torch.Size([2, 3, 3])
    """
    if any(is_batch(g) for g in graphs):
        raise ValueError("Cannot recursively batch graphs")

    # easy properties: just cat these together
    Z = torch.cat([g.Z for g in graphs])
    R = torch.cat([g.R for g in graphs])
    neighbour_offsets = torch.cat([g.neighbour_cell_offsets for g in graphs])

    # stack cells along a new batch dimension
    if not all_equal([has_cell(g) for g in graphs]):
        warnings.warn(
            "Attempting to batch a colleciton of graphs where only some "
            "have a defined unit cell. This may lead to unexpected results.",
            stacklevel=2,
        )
    cells = torch.stack([g.cell for g in graphs])

    # standard way to caculaute the batch and ptr properties
    batch = torch.cat(
        [torch.full_like(g.Z, fill_value=i) for i, g in enumerate(graphs)]
    )
    ptr = torch.tensor([0] + [g.Z.shape[0] for g in graphs]).cumsum(dim=0)

    # use the ptr to increment the neighbour index appropriately
    neighbour_list = torch.cat(
        [g.neighbour_list + ptr[i] for i, g in enumerate(graphs)], dim=1
    )

    properties: dict[PropertyKey, torch.Tensor] = {}
    # - per structure labels are concatenated along a new batch axis (0)
    for label in ["energy", "stress"]:
        if label in graphs[0].properties:
            properties[label] = torch.stack(
                [g.properties[label] for g in graphs]
            )

    # - per atom labels are concatenated along the first axis
    for key in ["forces", "local_energies"]:
        if key in graphs[0].properties:
            properties[key] = torch.cat([g.properties[key] for g in graphs])

    # - finally, add in the other stuff: this is a bit tricky
    #   since we need to try and infer whether these are per-atom
    #   or per-structure
    other = {"batch": batch, "ptr": ptr}

    for key in graphs[0].other:
        if key in ["batch", "ptr"]:
            continue

        # special handling for cutoff
        if key == "cutoff":
            cutoffs = [g.other["cutoff"] for g in graphs]
            if not all_equal(cutoffs):
                warnings.warn(
                    "Attempting to batch graphs with different cutoffs: "
                    f"{cutoffs}. Using the maximum cutoff.",
                    stacklevel=2,
                )
            other["cutoff"] = torch.tensor(cutoffs).max()
            continue

        # if all of the value tensors have the same number of entries
        # as atoms in the structure, we'll treat them as per-atom
        if all(is_local_property(g.other[key], g) for g in graphs):
            other[key] = torch.cat([g.other[key] for g in graphs])
        # otherwise, we'll assume they're per-structure
        else:
            other[key] = torch.stack([g.other[key] for g in graphs])

    return AtomicGraph(
        Z=Z,
        R=R,
        cell=cells,
        neighbour_list=neighbour_list,
        neighbour_cell_offsets=neighbour_offsets,
        properties=properties,
        other=other,
    )


def is_batch(graph: AtomicGraph) -> bool:
    """
    Does ``graph`` represent a batch of atomic graphs?

    Parameters
    ----------
    graph
        The graph to check.
    """

    return "batch" in graph.other and "ptr" in graph.other


############################### PROPERTIES ###############################


def number_of_atoms(graph: AtomicGraph) -> int:
    """
    Get the number of atoms in the ``graph``.
    """

    return graph.Z.shape[0]


def number_of_edges(graph: AtomicGraph) -> int:
    """
    Get the number of edges in the ``graph``.
    """

    return graph.neighbour_list.shape[1]


def has_cell(graph: AtomicGraph) -> bool:
    """
    Does ``graph`` represent a structure with a defined unit cell?
    """

    return not torch.allclose(graph.cell, torch.zeros_like(graph.cell))


def neighbour_vectors(graph: AtomicGraph) -> torch.Tensor:
    """
    Get the vector between each pair of atoms specified in the
    ``graph``'s ``"neighbour_list"`` property, respecting periodic
    boundary conditions where present.
    """

    # to simplify the logic below, we'll expand
    # a single graph into a batch of one
    if is_batch(graph):
        batch = graph.other["batch"]
        cell = graph.cell
    else:
        batch = torch.zeros_like(graph.Z)
        cell = graph.cell.unsqueeze(0)

    # avoid tuple de-structuring to keep torchscript happy
    i, j = graph.neighbour_list[0], graph.neighbour_list[1]  # (E,)
    cell_per_edge = cell[batch[i]]  # (E, 3, 3)
    distance_offsets = torch.einsum(
        "kl,klm->km",
        graph.neighbour_cell_offsets.to(cell_per_edge.dtype),
        cell_per_edge,
    )  # (E, 3)
    neighbour_positions = graph.R[j] + distance_offsets  # (E, 3)
    return neighbour_positions - graph.R[i]  # (E, 3)


def neighbour_distances(graph: AtomicGraph) -> torch.Tensor:
    """
    Get the distance between each pair of atoms specified in the
    ``graph``'s ``neighbour_list`` property, respecting periodic
    boundary conditions where present.
    """
    return torch.linalg.norm(neighbour_vectors(graph), dim=-1)


def number_of_structures(graph: AtomicGraph) -> int:
    """
    Get the number of structures in the ``graph``.
    """

    if not is_batch(graph):
        return 1
    return graph.other["ptr"].shape[0] - 1


def structure_sizes(batch: AtomicGraph) -> torch.Tensor:
    """
    Get the number of atoms in each structure in the ``batch``, of shape
    ``(S,)`` where ``S`` is the number of structures.

    Parameters
    ----------
    batch
        The batch to get the structure sizes for.

    Examples
    --------
    >>> len(graphs)
    3
    >>> [number_of_atoms(g) for g in graphs]
    [3, 4, 5]
    >>> structure_sizes(to_batch(graphs))
    tensor([3, 4, 5])
    """

    if not is_batch(batch):
        return torch.scalar_tensor(number_of_atoms(batch))

    return batch.other["ptr"][1:] - batch.other["ptr"][:-1]


def number_of_neighbours(
    graph: AtomicGraph,
    include_central_atom: bool = True,
) -> torch.Tensor:
    """
    Get a tensor, ``T``, of shape ``(N,)``, where ``N`` is the number of atoms
    in the ``graph``, such that ``T[i]`` gives the number of neighbours of atom
    ``i``. If ``include_central_atom`` is ``True``, then the central atom is
    included in the count.

    Parameters
    ----------
    graph
        The graph to get the number of neighbours for.
    include_central_atom
        Whether to include the central atom in the count.
    """

    return sum_over_neighbours(
        torch.ones_like(graph.neighbour_list[0]),
        graph,
    ) + int(include_central_atom)


def available_properties(graph: AtomicGraph) -> List[PropertyKey]:
    """Get the labels that are available on the ``graph``."""
    return [cast(PropertyKey, k) for k in graph.properties]


############################### ACTIONS ###############################


def is_local_property(x: torch.Tensor, graph: AtomicGraph) -> bool:
    """
    Is the property ``x`` local to each atom in the ``graph``?

    Parameters
    ----------
    x
        The property to check.
    graph
        The graph to check the property for.
    """

    return len(x.shape) > 0 and x.shape[0] == number_of_atoms(graph)


def trim_edges(graph: AtomicGraph, cutoff: float) -> AtomicGraph:
    """
    Return a new graph with edges trimmed to be no longer than the ``cutoff``.
    Leaves the original graph unchanged.

    Parameters
    ----------
    graph
        The graph to trim the edges of.
    cutoff
        The maximum distance between atoms to keep the edge.
    """

    existing_cutoff = graph.other.get(
        "cutoff", torch.tensor(float("inf"))
    ).item()
    if existing_cutoff < cutoff:
        warnings.warn(
            f"Graph already has a cutoff of {existing_cutoff} which is "
            f"less than the requested cutoff of {cutoff}.",
            stacklevel=2,
        )
        return graph
    elif existing_cutoff == cutoff:
        return graph

    distances = neighbour_distances(graph)
    mask = distances <= cutoff
    neighbour_list = graph.neighbour_list[:, mask]
    neighbour_cell_offsets = graph.neighbour_cell_offsets[mask, :]

    other = graph.other.copy()
    other["cutoff"] = torch.tensor(cutoff)

    # can't use _replace here due to TorchScript
    return AtomicGraph(
        Z=graph.Z,
        R=graph.R,
        cell=graph.cell,
        neighbour_list=neighbour_list,
        neighbour_cell_offsets=neighbour_cell_offsets,
        properties=graph.properties,
        other=other,
    )


def sum_over_neighbours(p: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    r"""
    Shape-preserving sum over neighbours of a per-edge property, :math:`p_{ij}`,
    to get a per-atom property, :math:`P_i`:

    .. math::
        P_i = \sum_{j \in \mathcal{N}_i} p_{ij}

    where:

    * :math:`\mathcal{N}_i` is the set of neighbours of atom :math:`i`.
    * :math:`p_{ij}` is the property of the edge between atoms :math:`i` and
      :math:`j`.
    * :math:`p` is of shape :code:`(E, ...)` and :math:`P` is of shape
      :code:`(N, ...)` where :math:`E` is the number of edges and :math:`N` is
      the number of atoms. :code:`...` denotes any number of additional
      dimensions, including none.
    * :math:`P_i` = 0 if :math:`|\mathcal{N}_i| = 0`.

    Parameters
    ----------
    p
        The per-edge property to sum.
    graph
        The graph to sum the property for.
    """

    N = number_of_atoms(graph)
    central_atoms = graph.neighbour_list[0]  # shape: (E,)

    # optimised implementations for common cases
    if p.dim() == 1:
        zeros = torch.zeros(N, dtype=p.dtype, device=p.device)
        return zeros.scatter_add(0, central_atoms, p)

    elif p.dim() == 2:
        C = p.shape[1]
        zeros = torch.zeros(N, C, dtype=p.dtype, device=p.device)
        return zeros.scatter_add(0, central_atoms.unsqueeze(1).expand(-1, C), p)

    shape = (N,) + p.shape[1:]
    zeros = torch.zeros(shape, dtype=p.dtype, device=p.device)

    if p.shape[0] == 0:
        # return all zeros if there are no atoms
        return zeros

    # create `index`, where index.shape = p.shape
    # and (index[e] == central_atoms[e]).all()
    ones = torch.ones_like(p)
    index = left_aligned_mul(ones, central_atoms).long()
    return zeros.scatter_add(0, index, p)


def sum_per_structure(x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    r"""
    Shape-preserving sum of a per-atom property, :math:`p`, to get a
    per-structure property, :math:`P`:

    If a single structure, containing ``N`` atoms, is used, then
    :math:`P = \sum_i p_i`, where:

    * :math:`p_i` is of shape ``(N, ...)``
    * :math:`P` is of shape ``(...)``
    * ``...`` denotes any
      number of additional dimensions, including ``None``.

    If a batch of ``S`` structures, containing a total of ``N`` atoms, is
    used, then :math:`P_k = \sum_{k \in K} p_k`, where:

    * :math:`K` is the collection of all atoms in structure :math:`k`
    * :math:`p_i` is of shape ``(N, ...)``
    * :math:`P` is of shape ``(S, ...)``
    * ``...`` denotes any
      number of additional dimensions, including ``None``.

    Parameters
    ----------
    x
        The per-atom property to sum.
    graph
        The graph to sum the property for.

    Examples
    --------
    Single graph case:

    >>> import torch
    >>> from ase.build import molecule
    >>> from graph_pes.atomic_graph import sum_per_structure, AtomicGraph
    >>> water = molecule("H2O")
    >>> graph = AtomicGraph.from_ase(water, cutoff=1.5)
    >>> # summing over a vector gives a scalar
    >>> sum_per_structure(torch.ones(3), graph)
    tensor(3.)
    >>> # summing over higher order tensors gives a tensor
    >>> sum_per_structure(torch.ones(3, 2, 3), graph).shape
    torch.Size([2, 3])

    Batch case:

    >>> import torch
    >>> from ase.build import molecule
    >>> from graph_pes.atomic_graph import sum_per_structure, AtomicGraph, to_batch
    >>> water = molecule("H2O")
    >>> graph = AtomicGraph.from_ase(water, cutoff=1.5)
    >>> batch = to_batch([graph, graph])
    >>> batch
    AtomicGraphBatch(structures: 2, atoms: 6, edges: 8, has_cell: False)
    >>> # summing over a vector gives a tensor
    >>> sum_per_structure(torch.ones(6), graph)
    tensor([3., 3.])
    >>> # summing over higher order tensors gives a tensor
    >>> sum_per_structure(torch.ones(6, 3, 4), graph).shape
    torch.Size([2, 3, 4])
    """  # noqa: E501

    if is_batch(graph):
        batch = graph.other["batch"]
        shape = (number_of_structures(graph),) + x.shape[1:]
        zeros = torch.zeros(shape, dtype=x.dtype, device=x.device)
        return zeros.scatter_add(0, batch, x)
    else:
        return x.sum(dim=0)


def index_over_neighbours(x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    """
    Index a per-atom property, :math:`x`, over the neighbours of each atom in
    the ``graph``.
    """
    return x[graph.neighbour_list[1]]


def divide_per_atom(x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    r"""
    Divide a per-structure property, :math:`X`, by the number of atoms in each
    structure to get a per-atom property, :math:`x`:

    .. math::

        x_i = \frac{X_k}{N_k}

    where:

    * :math:`X` is of shape ``(S, ...)``
    * :math:`x` is of shape ``(N, ...)``
    * :math:`S` is the number of structures
    * :math:`N` is the number of atoms
    """
    return left_aligned_div(x, structure_sizes(graph))



from __future__ import annotations

import warnings

warnings.filterwarnings(
    "ignore",
    module="e3nn",
    message=".*you are using `torch.load`.*",
)

import pathlib

import torch

from graph_pes.graph_pes_model import GraphPESModel

from .addition import AdditionModel
from .e3nn.mace import MACE, ZEmbeddingMACE
from .e3nn.nequip import NequIP, ZEmbeddingNequIP
from .offsets import FixedOffset, LearnableOffset
from .painn import PaiNN
from .pairwise import (
    LennardJones,
    LennardJonesMixture,
    Morse,
    PairPotential,
    SmoothedPairPotential,
    ZBLCoreRepulsion,
)
from .schnet import SchNet
from .tensornet import TensorNet

# TODO get rid of this all business
__all__ = [
    "AdditionModel",
    "FixedOffset",
    "LearnableOffset",
    "LennardJones",
    "LennardJonesMixture",
    "MACE",
    "Morse",
    "NequIP",
    "PaiNN",
    "PairPotential",
    "SchNet",
    "SmoothedPairPotential",
    "TensorNet",
    "ZBLCoreRepulsion",
    "ZEmbeddingMACE",
    "ZEmbeddingNequIP",
]

# TODO: nicer way to do this?
ALL_MODELS: list[type[GraphPESModel]] = [
    globals()[model]
    for model in __all__
    if model
    not in [
        "FixedOffset",
        "LearnableOffset",
        "AdditionModel",
        "PairPotential",
        "SmoothedPairPotential",
    ]
]


def load_model(
    path: str | pathlib.Path,
) -> GraphPESModel:
    """
    Load a model from a file.

    Parameters
    ----------
    path
        The path to the file.

    Returns
    -------
    GraphPESModel
        The model.

    Examples
    --------

    Use this function to load an existing model for further training using
    ``graph-pes-train``:

    .. code-block:: yaml

        model:
            graph_pes.models.load_model:
                path: path/to/model.pt

    To account for some new energy offset in your training data, you could do
    something like this:
    (see also :func:`~graph_pes.models.load_model_component`)

    .. code-block:: yaml

        model:
            # add an offset to an existing model before fine-tuning
            offset:
                graph_pes.models.LearnableOffset: {}
            many-body:
                graph_pes.models.load_model:
                    path: path/to/model.pt
    """
    path = pathlib.Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Could not find model at {path}")

    return torch.load(path)


def load_model_component(path: str | pathlib.Path, key: str) -> GraphPESModel:
    """
    Load a component from an :class:`~graph_pes.models.AdditionModel`.

    Parameters
    ----------
    path
        The path to the file.
    key
        The key to load.

    Returns
    -------
    GraphPESModel
        The component.

    Examples
    --------

    Train on data with a new energy offset:

    .. code-block:: yaml

        model:
            offset:
                graph_pes.models.LearnableOffset: {}
            many-body:
                graph_pes.models.load_model_component:
                    path: path/to/model.pt
                    key: many-body
    """

    base_model = load_model(path)
    if not isinstance(base_model, AdditionModel):
        raise ValueError(
            f"Expected to load an AdditionModel, got {type(base_model)}"
        )

    return base_model[key]
